// Generated from player/lua/fzy.lua

"--[[ The MIT License (MIT)\n"
"\n"
"Copyright (c) 2020 Seth Warn\n"
"\n"
"Permission is hereby granted, free of charge, to any person obtaining a copy\n"
"of this software and associated documentation files (the \"Software\"), to deal\n"
"in the Software without restriction, including without limitation the rights\n"
"to use, copy, modify, merge, publish, distribute, sublicense, and/or sell\n"
"copies of the Software, and to permit persons to whom the Software is\n"
"furnished to do so, subject to the following conditions:\n"
"\n"
"The above copyright notice and this permission notice shall be included in\n"
"all copies or substantial portions of the Software.\n"
"\n"
"THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\n"
"IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\n"
"FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\n"
"AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\n"
"LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\n"
"OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN\n"
"THE SOFTWARE. ]]\n"
"\n"
"-- The lua implementation of the fzy string matching algorithm\n"
"\n"
"local SCORE_GAP_LEADING = -0.005\n"
"local SCORE_GAP_TRAILING = -0.005\n"
"local SCORE_GAP_INNER = -0.01\n"
"local SCORE_MATCH_CONSECUTIVE = 1.0\n"
"local SCORE_MATCH_SLASH = 0.9\n"
"local SCORE_MATCH_WORD = 0.8\n"
"local SCORE_MATCH_CAPITAL = 0.7\n"
"local SCORE_MATCH_DOT = 0.6\n"
"local SCORE_MAX = math.huge\n"
"local SCORE_MIN = -math.huge\n"
"local MATCH_MAX_LENGTH = 1024\n"
"\n"
"local fzy = {}\n"
"\n"
"-- Check if \140needle\140 is a subsequence of the \140haystack\140.\n"
"--\n"
"-- Usually called before \140score\140 or \140positions\140.\n"
"--\n"
"-- Args:\n"
"--   needle (string)\n"
"--   haystack (string)\n"
"--   case_sensitive (bool, optional): defaults to false\n"
"--\n"
"-- Returns:\n"
"--   bool\n"
"function fzy.has_match(needle, haystack, case_sensitive)\n"
"  if not case_sensitive then\n"
"    needle = string.lower(needle)\n"
"    haystack = string.lower(haystack)\n"
"  end\n"
"\n"
"  local j = 1\n"
"  for i = 1, string.len(needle) do\n"
"    j = string.find(haystack, needle:sub(i, i), j, true)\n"
"    if not j then\n"
"      return false\n"
"    else\n"
"      j = j + 1\n"
"    end\n"
"  end\n"
"\n"
"  return true\n"
"end\n"
"\n"
"local function is_lower(c)\n"
"  return c:match(\"%l\")\n"
"end\n"
"\n"
"local function is_upper(c)\n"
"  return c:match(\"%u\")\n"
"end\n"
"\n"
"local function precompute_bonus(haystack)\n"
"  local match_bonus = {}\n"
"\n"
"  local last_char = \"/\"\n"
"  for i = 1, string.len(haystack) do\n"
"    local this_char = haystack:sub(i, i)\n"
"    if last_char == \"/\" or last_char == \"\\\\\" then\n"
"      match_bonus[i] = SCORE_MATCH_SLASH\n"
"    elseif last_char == \"-\" or last_char == \"_\" or last_char == \" \" then\n"
"      match_bonus[i] = SCORE_MATCH_WORD\n"
"    elseif last_char == \".\" then\n"
"      match_bonus[i] = SCORE_MATCH_DOT\n"
"    elseif is_lower(last_char) and is_upper(this_char) then\n"
"      match_bonus[i] = SCORE_MATCH_CAPITAL\n"
"    else\n"
"      match_bonus[i] = 0\n"
"    end\n"
"\n"
"    last_char = this_char\n"
"  end\n"
"\n"
"  return match_bonus\n"
"end\n"
"\n"
"local function compute(needle, haystack, D, M, case_sensitive)\n"
"  -- Note that the match bonuses must be computed before the arguments are\n"
"  -- converted to lowercase, since there are bonuses for camelCase.\n"
"  local match_bonus = precompute_bonus(haystack)\n"
"  local n = string.len(needle)\n"
"  local m = string.len(haystack)\n"
"\n"
"  if not case_sensitive then\n"
"    needle = string.lower(needle)\n"
"    haystack = string.lower(haystack)\n"
"  end\n"
"\n"
"  -- Because lua only grants access to chars through substring extraction,\n"
"  -- get all the characters from the haystack once now, to reuse below.\n"
"  local haystack_chars = {}\n"
"  for i = 1, m do\n"
"    haystack_chars[i] = haystack:sub(i, i)\n"
"  end\n"
"\n"
"  for i = 1, n do\n"
"    D[i] = {}\n"
"    M[i] = {}\n"
"\n"
"    local prev_score = SCORE_MIN\n"
"    local gap_score = i == n and SCORE_GAP_TRAILING or SCORE_GAP_INNER\n"
"    local needle_char = needle:sub(i, i)\n"
"\n"
"    for j = 1, m do\n"
"      if needle_char == haystack_chars[j] then\n"
"        local score = SCORE_MIN\n"
"        if i == 1 then\n"
"          score = ((j - 1) * SCORE_GAP_LEADING) + match_bonus[j]\n"
"        elseif j > 1 then\n"
"          local a = M[i - 1][j - 1] + match_bonus[j]\n"
"          local b = D[i - 1][j - 1] + SCORE_MATCH_CONSECUTIVE\n"
"          score = math.max(a, b)\n"
"        end\n"
"        D[i][j] = score\n"
"        prev_score = math.max(score, prev_score + gap_score)\n"
"        M[i][j] = prev_score\n"
"      else\n"
"        D[i][j] = SCORE_MIN\n"
"        prev_score = prev_score + gap_score\n"
"        M[i][j] = prev_score\n"
"      end\n"
"    end\n"
"  end\n"
"end\n"
"\n"
"-- Compute a matching score.\n"
"--\n"
"-- Args:\n"
"--   needle (string): must be a subequence of \140haystack\140, or the result is\n"
"--     undefined.\n"
"--   haystack (string)\n"
"--   case_sensitive (bool, optional): defaults to false\n"
"--\n"
"-- Returns:\n"
"--   number: higher scores indicate better matches. See also \140get_score_min\140\n"
"--     and \140get_score_max\140.\n"
"function fzy.score(needle, haystack, case_sensitive)\n"
"  local n = string.len(needle)\n"
"  local m = string.len(haystack)\n"
"\n"
"  if n == 0 or m == 0 or m > MATCH_MAX_LENGTH or n > m then\n"
"    return SCORE_MIN\n"
"  elseif n == m then\n"
"    return SCORE_MAX\n"
"  else\n"
"    local D = {}\n"
"    local M = {}\n"
"    compute(needle, haystack, D, M, case_sensitive)\n"
"    return M[n][m]\n"
"  end\n"
"end\n"
"\n"
"-- Compute the locations where fzy matches a string.\n"
"--\n"
"-- Determine where each character of the \140needle\140 is matched to the \140haystack\140\n"
"-- in the optimal match.\n"
"--\n"
"-- Args:\n"
"--   needle (string): must be a subequence of \140haystack\140, or the result is\n"
"--     undefined.\n"
"--   haystack (string)\n"
"--   case_sensitive (bool, optional): defaults to false\n"
"--\n"
"-- Returns:\n"
"--   {int,...}: indices, where \140indices[n]\140 is the location of the \140n\140th\n"
"--     character of \140needle\140 in \140haystack\140.\n"
"--   number: the same matching score returned by \140score\140\n"
"function fzy.positions(needle, haystack, case_sensitive)\n"
"  local n = string.len(needle)\n"
"  local m = string.len(haystack)\n"
"\n"
"  if n == 0 or m == 0 or m > MATCH_MAX_LENGTH or n > m then\n"
"    return {}, SCORE_MIN\n"
"  elseif n == m then\n"
"    local consecutive = {}\n"
"    for i = 1, n do\n"
"      consecutive[i] = i\n"
"    end\n"
"    return consecutive, SCORE_MAX\n"
"  end\n"
"\n"
"  local D = {}\n"
"  local M = {}\n"
"  compute(needle, haystack, D, M, case_sensitive)\n"
"\n"
"  local positions = {}\n"
"  local match_required = false\n"
"  local j = m\n"
"  for i = n, 1, -1 do\n"
"    while j >= 1 do\n"
"      if D[i][j] ~= SCORE_MIN and (match_required or D[i][j] == M[i][j]) then\n"
"        match_required = (i ~= 1) and (j ~= 1) and (\n"
"        M[i][j] == D[i - 1][j - 1] + SCORE_MATCH_CONSECUTIVE)\n"
"        positions[i] = j\n"
"        j = j - 1\n"
"        break\n"
"      else\n"
"        j = j - 1\n"
"      end\n"
"    end\n"
"  end\n"
"\n"
"  return positions, M[n][m]\n"
"end\n"
"\n"
"-- Apply \140has_match\140 and \140positions\140 to an array of haystacks.\n"
"--\n"
"-- Args:\n"
"--   needle (string)\n"
"--   haystack ({string, ...})\n"
"--   case_sensitive (bool, optional): defaults to false\n"
"--\n"
"-- Returns:\n"
"--   {{idx, positions, score}, ...}: an array with one entry per matching line\n"
"--     in \140haystacks\140, each entry giving the index of the line in \140haystacks\140\n"
"--     as well as the equivalent to the return value of \140positions\140 for that\n"
"--     line.\n"
"function fzy.filter(needle, haystacks, case_sensitive)\n"
"  local result = {}\n"
"\n"
"  for i, line in ipairs(haystacks) do\n"
"    if fzy.has_match(needle, line, case_sensitive) then\n"
"      local p, s = fzy.positions(needle, line, case_sensitive)\n"
"      table.insert(result, {i, p, s})\n"
"    end\n"
"  end\n"
"\n"
"  return result\n"
"end\n"
"\n"
"-- The lowest value returned by \140score\140.\n"
"--\n"
"-- In two special cases:\n"
"--  - an empty \140needle\140, or\n"
"--  - a \140needle\140 or \140haystack\140 larger than than \140get_max_length\140,\n"
"-- the \140score\140 function will return this exact value, which can be used as a\n"
"-- sentinel. This is the lowest possible score.\n"
"function fzy.get_score_min()\n"
"  return SCORE_MIN\n"
"end\n"
"\n"
"-- The score returned for exact matches. This is the highest possible score.\n"
"function fzy.get_score_max()\n"
"  return SCORE_MAX\n"
"end\n"
"\n"
"-- The maximum size for which \140fzy\140 will evaluate scores.\n"
"function fzy.get_max_length()\n"
"  return MATCH_MAX_LENGTH\n"
"end\n"
"\n"
"-- The minimum score returned for normal matches.\n"
"--\n"
"-- For matches that don't return \140get_score_min\140, their score will be greater\n"
"-- than than this value.\n"
"function fzy.get_score_floor()\n"
"  return MATCH_MAX_LENGTH * SCORE_GAP_INNER\n"
"end\n"
"\n"
"-- The maximum score for non-exact matches.\n"
"--\n"
"-- For matches that don't return \140get_score_max\140, their score will be less than\n"
"-- this value.\n"
"function fzy.get_score_ceiling()\n"
"  return MATCH_MAX_LENGTH * SCORE_MATCH_CONSECUTIVE\n"
"end\n"
"\n"
"-- The name of the currently-running implmenetation, \"lua\" or \"native\".\n"
"function fzy.get_implementation_name()\n"
"  return \"lua\"\n"
"end\n"
"\n"
"return fzy\n"
